use anyhow::{
    Context,
    Result,
};
use chrono::{
    DateTime,
    Utc,
};
use crate::utils::{
    urlencode,
};
use sqlx::{
    FromRow,
    SqlitePool,
    sqlite::{
        SqlitePoolOptions,
        SqliteQueryResult,
    },
};

#[derive(Clone)]
pub struct Storage {
    pub db: SqlitePool,
}

impl Storage {
    pub async fn connect(db_connection_string: &str) -> Result<Self> {
        let db = SqlitePoolOptions::new()
            .connect(db_connection_string)
            .await
            .context("Failed to connect to database")?;

        Ok(Self {
            db,
        })
    }

    pub async fn set_flake(&self, flake_row: FlakeRow) -> Result<SqliteQueryResult> {
        sqlx::query("INSERT INTO flakes (flake_uri, tracker_track, tracker_last_scanned)
            VALUES (?, ?, ?)
            ON CONFLICT(flake_uri) DO UPDATE SET
                tracker_track=excluded.tracker_track,
                tracker_last_scanned=excluded.tracker_last_scanned
            ")
            .bind(&flake_row.flake_uri)
            .bind(&flake_row.tracker_track)
            .bind(&flake_row.tracker_last_scanned)
            .execute(&self.db)
            .await
            .context("Failed to execute database query")
    }

    pub async fn set_flake_exist(&self, flake_row: FlakeRow) -> Result<SqliteQueryResult> {
        sqlx::query("INSERT INTO flakes (flake_uri, tracker_track)
            VALUES (?, FALSE)
            ON CONFLICT(flake_uri) DO NOTHING
            ")
            .bind(&flake_row.flake_uri)
            .execute(&self.db)
            .await
            .context("Failed to execute database query")
    }

    pub async fn flakes(&self) -> Result<Vec<FlakeRow>> {
        sqlx::query_as("
            SELECT
                flake_uri,
                tracker_track,
                tracker_last_scanned
            FROM flakes
            ORDER BY flake_uri
            ")
            .fetch_all(&self.db)
            .await
            .context("Failed to fetch data from database")
    }

    pub async fn set_revision(&self, revision_row: RevisionRow) -> Result<SqliteQueryResult> {
        sqlx::query("INSERT INTO revisions (revision_uri, flake_uri, nix_store_path, nar_hash, last_modified)
            VALUES (?, ?, ?, ?, ?)
            ON CONFLICT(revision_uri) DO UPDATE SET
                flake_uri=excluded.flake_uri,
                nix_store_path=excluded.nix_store_path,
                nar_hash=excluded.nar_hash,
                last_modified=excluded.last_modified
            ")
            .bind(&revision_row.revision_uri)
            .bind(&revision_row.flake_uri)
            .bind(&revision_row.nix_store_path)
            .bind(&revision_row.nar_hash)
            .bind(&revision_row.last_modified)
            .execute(&self.db)
            .await
            .context("Failed to execute database query")
    }

    pub async fn set_revision_exist(&self, revision_row: RevisionRow) -> Result<SqliteQueryResult> {
        sqlx::query("INSERT INTO revisions (revision_uri, flake_uri)
            VALUES (?, ?)
            ON CONFLICT(revision_uri) DO NOTHING
            ")
            .bind(&revision_row.revision_uri)
            .bind(&revision_row.flake_uri)
            .execute(&self.db)
            .await
            .context("Failed to execute database query")
    }

    pub async fn revisions_from_flake(&self, uri: &str) -> Result<Vec<RevisionRow>> {
        sqlx::query_as("
            SELECT
                revision_uri,
                flake_uri,
                nix_store_path,
                nar_hash,
                last_modified
            FROM revisions
            WHERE flake_uri = ?
            ORDER BY last_modified DESC
            ")
            .bind(&uri)
            .fetch_all(&self.db)
            .await
            .context("Failed to fetch data from database")
    }

    pub async fn set_input(&self, input_row: InputRow) -> Result<SqliteQueryResult> {
        sqlx::query("INSERT INTO inputs (revision_uri, input_name, locked_revision_uri, locked_flake_uri, locked_nar_hash, last_modified)
            VALUES (?, ?, ?, ?, ?, ?)
            ON CONFLICT(revision_uri, input_name) DO UPDATE SET
                locked_revision_uri=excluded.locked_revision_uri,
                locked_flake_uri=excluded.locked_flake_uri,
                locked_nar_hash=excluded.locked_nar_hash,
                last_modified=excluded.last_modified
            ")
            .bind(input_row.revision_uri)
            .bind(input_row.input_name)
            .bind(input_row.locked_revision_uri)
            .bind(input_row.locked_flake_uri)
            .bind(input_row.locked_nar_hash)
            .bind(input_row.last_modified)
            .execute(&self.db)
            .await
            .context("Failed to execute database query")

    }

    pub async fn inputs_for_revision(&self, revision_uri: &str) -> Result<Vec<InputRow>> {
        sqlx::query_as("
            SELECT
                revision_uri,
                input_name,
                locked_revision_uri,
                locked_flake_uri,
                locked_nar_hash,
                last_modified
            FROM inputs
            WHERE revision_uri = ?
            ORDER BY input_name
            ")
            .bind(&revision_uri)
            .fetch_all(&self.db)
            .await
            .context("Failed to fetch data from database")
    }

    pub async fn input_of_for_revision(&self, revision_uri: &str) -> Result<Vec<InputRow>> {
        sqlx::query_as("
            SELECT
                revision_uri,
                input_name,
                locked_revision_uri,
                locked_flake_uri,
                locked_nar_hash,
                last_modified
            FROM inputs
            WHERE locked_revision_uri = ?
            ORDER BY input_name
            ")
            .bind(&revision_uri)
            .fetch_all(&self.db)
            .await
            .context("Failed to fetch data from database")
    }

    pub async fn current_inputs_for_flake(&self, flake_uri: &str) -> Result<Vec<InputRow>> {
        sqlx::query_as("

            SELECT
                revisions.revision_uri,
                inputs.input_name,
                inputs.locked_revision_uri,
                inputs.locked_flake_uri,
                inputs.locked_nar_hash,
                inputs.last_modified,
                MAX(revisions.last_modified)
            FROM
                revisions
            LEFT JOIN
                inputs
            ON
                revisions.revision_uri = inputs.revision_uri
            WHERE
                revisions.flake_uri = ?
            GROUP BY
                inputs.input_name
            ")
            .bind(&flake_uri)
            .fetch_all(&self.db)
            .await
            .context("Failed to fetch data from database")
    }
}

#[derive(FromRow)]
pub struct RevisionRow {
    pub revision_uri: String,
    pub flake_uri: Option<String>,
    pub nix_store_path: Option<String>,
    pub nar_hash: Option<String>,
    pub last_modified: Option<i64>,
}

impl RevisionRow {
    pub fn revision_link(&self) -> String {
        format!("/r/{}", urlencode(&self.revision_uri))
    }

    pub fn flake_link(&self) -> String {
        match &self.flake_uri {
            Some(flake_uri) => format!("/f/{}", urlencode(&flake_uri)),
            None => String::from("#"),
        }
    }

    pub fn last_modified_time(&self) -> Option<DateTime<Utc>> {
        match &self.last_modified {
            Some(last_modified) => DateTime::from_timestamp(last_modified.clone(), 0),
            None => None,
        }
    }
}

#[derive(FromRow)]
pub struct InputRow {
    pub revision_uri: String,
    pub input_name: String,
    pub locked_revision_uri: Option<String>,
    pub locked_flake_uri: Option<String>,
    pub locked_nar_hash: Option<String>,
    pub last_modified: Option<i64>,
}

impl InputRow {
    pub fn revision_link(&self) -> String {
        format!("/r/{}", urlencode(&self.revision_uri))
    }

    pub fn locked_revision_link(&self) -> String {
        match &self.locked_revision_uri {
            Some(locked_revision_uri) => format!("/r/{}", urlencode(&locked_revision_uri)),
            None => String::from("#"),
        }
    }

    pub fn locked_flake_link(&self) -> String {
        match &self.locked_flake_uri {
            Some(locked_flake_uri) => format!("/f/{}", urlencode(&locked_flake_uri)),
            None => String::from("#"),
        }
    }
}

#[derive(FromRow)]
pub struct FlakeRow {
    pub flake_uri: String,
    pub tracker_track: bool,
    pub tracker_last_scanned: Option<i64>,
}

impl FlakeRow {
    pub fn flake_link(&self ) -> String {
        format!("/f/{}", urlencode(&self.flake_uri))
    }
}

